// TypeTest4.cpp
// Copyright (C) 2019 Katayama Hirofumi MZ <katayama.hirofumi.mz@gmail.com>
// This file is public domain software.
#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include <tchar.h>
#include <cmath>

#define FILENAME "arial.ttf"
#define FONTNAME "Arial"
#define LFHEIGHT -50
#define LFWIDTH 0

#ifndef NO_FREETYPE
    #include <ft2build.h>
    #include FT_FREETYPE_H
    #include FT_GLYPH_H
    #include FT_TRUETYPE_TABLES_H
    FT_Library g_library;
    FT_Face g_face;
#endif

static const CHAR g_szText[] = "ABCabc,";
#ifndef NO_FREETYPE
static const FT_Matrix identityMat = {(1 << 16), 0, 0, (1 << 16)};
#endif

const char *GetFontPath(const char *name)
{
    static char path[MAX_PATH];
    GetWindowsDirectoryA(path, MAX_PATH);
    lstrcat(path, "\\Fonts\\");
    lstrcat(path, name);
    return path;
}

typedef struct BITMAPINFODX
{
    BITMAPINFOHEADER    bmiHeader;
    RGBQUAD             bmiColors[256];
} BITMAPINFODX;

BOOL SaveBitmapToFile(LPCTSTR pszFileName, HBITMAP hbm)
{
    BOOL f;
    BITMAPFILEHEADER bf;
    BITMAPINFODX bi;
    BITMAPINFOHEADER *pbmih;
    DWORD cb;
    DWORD cColors, cbColors;
    HDC hDC;
    HANDLE hFile;
    LPVOID pBits;
    BITMAP bm;

    if (!GetObject(hbm, sizeof(BITMAP), &bm))
        return FALSE;

    pbmih = &bi.bmiHeader;
    ZeroMemory(pbmih, sizeof(BITMAPINFOHEADER));
    pbmih->biSize             = sizeof(BITMAPINFOHEADER);
    pbmih->biWidth            = bm.bmWidth;
    pbmih->biHeight           = bm.bmHeight;
    pbmih->biPlanes           = 1;
    pbmih->biBitCount         = bm.bmBitsPixel;
    pbmih->biCompression      = BI_RGB;
    pbmih->biSizeImage        = bm.bmWidthBytes * bm.bmHeight;

    if (bm.bmBitsPixel < 16)
        cColors = 1 << bm.bmBitsPixel;
    else
        cColors = 0;
    cbColors = cColors * sizeof(RGBQUAD);

    bf.bfType = 0x4d42;
    bf.bfReserved1 = 0;
    bf.bfReserved2 = 0;
    cb = sizeof(BITMAPFILEHEADER) + pbmih->biSize + cbColors;
    bf.bfOffBits = cb;
    bf.bfSize = cb + pbmih->biSizeImage;

    pBits = HeapAlloc(GetProcessHeap(), 0, pbmih->biSizeImage);
    if (pBits == NULL)
        return FALSE;

    f = FALSE;
    hDC = GetDC(NULL);
    if (hDC != NULL)
    {
        if (GetDIBits(hDC, hbm, 0, bm.bmHeight, pBits, (BITMAPINFO*)&bi,
            DIB_RGB_COLORS))
        {
            hFile = CreateFile(pszFileName, GENERIC_WRITE, FILE_SHARE_READ, NULL,
                               CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL |
                               FILE_FLAG_WRITE_THROUGH, NULL);
            if (hFile != INVALID_HANDLE_VALUE)
            {
                f = WriteFile(hFile, &bf, sizeof(BITMAPFILEHEADER), &cb, NULL) &&
                    WriteFile(hFile, &bi, sizeof(BITMAPINFOHEADER), &cb, NULL) &&
                    WriteFile(hFile, bi.bmiColors, cbColors, &cb, NULL) &&
                    WriteFile(hFile, pBits, pbmih->biSizeImage, &cb, NULL);
                CloseHandle(hFile);

                if (!f)
                    DeleteFile(pszFileName);
            }
        }
        ReleaseDC(NULL, hDC);
    }

    HeapFree(GetProcessHeap(), 0, pBits);
    return f;
}

void InitXForm(HWND hwnd)
{
    SetDlgItemText(hwnd, edt1, TEXT("1"));
    SetDlgItemText(hwnd, edt2, TEXT("0"));
    SetDlgItemText(hwnd, edt3, TEXT("0"));
    SetDlgItemText(hwnd, edt4, TEXT("1"));
    SetDlgItemText(hwnd, edt5, TEXT("0"));
    SetDlgItemText(hwnd, edt6, TEXT("0"));
}

#ifndef NO_FREETYPE
    HBITMAP ConvertToGrayscaleBitmap(FT_Bitmap bitmap)
    {
        HBITMAP hbm;
        BITMAPINFODX bi;
        
        ZeroMemory(&bi, sizeof(bi));
        bi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
        bi.bmiHeader.biWidth = bitmap.width;
        bi.bmiHeader.biHeight = -bitmap.rows;
        bi.bmiHeader.biPlanes = 1;
        bi.bmiHeader.biBitCount = 8;
        bi.bmiHeader.biCompression = BI_RGB;
        
        for (INT i = 0; i < 256; ++i)
        {
            bi.bmiColors[i].rgbBlue = (BYTE)i;
            bi.bmiColors[i].rgbGreen = (BYTE)i;
            bi.bmiColors[i].rgbRed = (BYTE)i;
            bi.bmiColors[i].rgbReserved = 0;
        }

        LPVOID pvBits;
        hbm = CreateDIBSection(NULL, (BITMAPINFO *)&bi, DIB_RGB_COLORS, &pvBits, NULL, 0);
        if (hbm == NULL)
            return NULL;

        BITMAP bm;
        GetObject(hbm, sizeof(bm), &bm);

        LPBYTE pbBits = (LPBYTE)pvBits;
        for (INT y = 0; y < bitmap.rows; ++y)
        {
            CopyMemory(&pbBits[y * bm.bmWidthBytes], 
                       &bitmap.buffer[y * bitmap.pitch],
                       bitmap.width);
        }

        return hbm;
    }

    FT_GlyphSlot LoadGlyphSlot(WCHAR charcode)
    {
        if (FT_Error error = FT_Load_Char(g_face, charcode, 0))
        {
            MessageBoxA(NULL, "FT_Load_Char", NULL, MB_ICONERROR);
            return NULL;
        }

        if (FT_Error error = FT_Render_Glyph(g_face->glyph, FT_RENDER_MODE_NORMAL))
        {
            MessageBoxA(NULL, "FT_Render_Glyph", NULL, MB_ICONERROR);
            return NULL;
        }

        return g_face->glyph;
    }

    VOID FASTCALL IntEscapeMatrix(FT_Matrix *pmat, LONG lfEscapement)
    {
        double radian = (lfEscapement * M_PI) / 180 / 10;

        pmat->xx = (FT_Fixed)(cos(radian) * (1 << 16));
        pmat->xy = (FT_Fixed)(-sin(radian) * (1 << 16));
        pmat->yx = (FT_Fixed)(sin(radian) * (1 << 16));
        pmat->yy = (FT_Fixed)(cos(radian) * (1 << 16));
    }

    VOID FASTCALL IntWidthMatrix(FT_Face face, FT_Matrix *pmat, LONG lfWidth)
    {
        LONG tmAveCharWidth;
        TT_OS2 *pOS2;
        FT_Fixed XScale;

        *pmat = identityMat;

        if (lfWidth == 0)
            return;

        //ASSERT_FREETYPE_LOCK_HELD();
        pOS2 = (TT_OS2 *)FT_Get_Sfnt_Table(face, FT_SFNT_OS2);
        if (!pOS2)
            return;

        XScale = face->size->metrics.x_scale;
        tmAveCharWidth = (FT_MulFix(pOS2->xAvgCharWidth, XScale) + 32) >> 6;
        if (tmAveCharWidth == 0)
        {
            tmAveCharWidth = 1;
        }

        if (lfWidth == tmAveCharWidth)
            return;

        pmat->xx = (FT_Fixed)((1 << 16) * lfWidth / tmAveCharWidth);
        pmat->xy = 0;
        pmat->yx = 0;
        pmat->yy = (FT_Fixed)(1 << 16);
    }
#endif

BOOL OnInitDialog(HWND hwnd, HWND hwndFocus, LPARAM lParam)
{
    HWND hCmb1 = GetDlgItem(hwnd, cmb1);
    HWND hCmb2 = GetDlgItem(hwnd, cmb2);
    HWND hCmb3 = GetDlgItem(hwnd, cmb3);
    HWND hCmb4 = GetDlgItem(hwnd, cmb4);

    ComboBox_AddString(hCmb1, TEXT("0"));
    ComboBox_AddString(hCmb1, TEXT("900"));
    ComboBox_AddString(hCmb1, TEXT("1800"));
    ComboBox_AddString(hCmb1, TEXT("2700"));
    ComboBox_SetCurSel(hCmb1, 0);

    ComboBox_AddString(hCmb2, TEXT("TA_TOP"));
    ComboBox_AddString(hCmb2, TEXT("TA_BASELINE"));
    ComboBox_AddString(hCmb2, TEXT("TA_BOTTOM"));
    ComboBox_SetCurSel(hCmb2, 0);

    ComboBox_AddString(hCmb3, TEXT("1:1"));
    ComboBox_AddString(hCmb3, TEXT("2:3"));
    ComboBox_AddString(hCmb3, TEXT("3:2"));
    ComboBox_SetCurSel(hCmb3, 0);

    ComboBox_AddString(hCmb4, TEXT("0"));
    ComboBox_AddString(hCmb4, TEXT("+80"));
    ComboBox_AddString(hCmb4, TEXT("-80"));
    ComboBox_SetCurSel(hCmb4, 0);

    InitXForm(hwnd);
    CheckDlgButton(hwnd, chx1, BST_CHECKED);

#ifdef NO_FREETYPE
    ShowWindow(GetDlgItem(hwnd, chx2), SW_HIDE);
#endif

    return TRUE;
}

void OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify)
{
    switch (id)
    {
    case IDOK:
    case IDCANCEL:
        EndDialog(hwnd, id);
        break;
    case cmb1:
    case cmb2:
    case cmb3:
    case cmb4:
        if (codeNotify == CBN_EDITCHANGE || codeNotify == CBN_SELCHANGE)
        {
            InvalidateRect(hwnd, NULL, TRUE);
        }
        break;
    case chx1:
        if (codeNotify == BN_CLICKED)
        {
            InvalidateRect(hwnd, NULL, TRUE);
            if (IsDlgButtonChecked(hwnd, chx1) == BST_CHECKED)
            {
                EnableWindow(GetDlgItem(hwnd, edt1), TRUE);
                EnableWindow(GetDlgItem(hwnd, edt2), TRUE);
                EnableWindow(GetDlgItem(hwnd, edt3), TRUE);
                EnableWindow(GetDlgItem(hwnd, edt4), TRUE);
                EnableWindow(GetDlgItem(hwnd, edt5), TRUE);
                EnableWindow(GetDlgItem(hwnd, edt6), TRUE);
            }
            else
            {
                EnableWindow(GetDlgItem(hwnd, edt1), FALSE);
                EnableWindow(GetDlgItem(hwnd, edt2), FALSE);
                EnableWindow(GetDlgItem(hwnd, edt3), FALSE);
                EnableWindow(GetDlgItem(hwnd, edt4), FALSE);
                EnableWindow(GetDlgItem(hwnd, edt5), FALSE);
                EnableWindow(GetDlgItem(hwnd, edt6), FALSE);
                InitXForm(hwnd);
            }
        }
        break;
    case chx2:
        if (codeNotify == BN_CLICKED)
        {
            InvalidateRect(hwnd, NULL, TRUE);
        }
        break;
    case edt1:
    case edt2:
    case edt3:
    case edt4:
    case edt5:
    case edt6:
        if (codeNotify == EN_CHANGE)
        {
            if (IsDlgButtonChecked(hwnd, chx1) == BST_CHECKED)
            {
                InvalidateRect(hwnd, NULL, TRUE);
            }
        }
        break;
    }
}

FLOAT GetDlgItemFloat(HWND hwnd, UINT nCtrlID)
{
    CHAR szText[64];
    GetDlgItemTextA(hwnd, nCtrlID, szText, ARRAYSIZE(szText));
    return atof(szText);
}

void OnWinDraw(HWND hwnd, HDC hDC)
{
    RECT rc;
    GetClientRect(hwnd, &rc);

    HWND hCmb3 = GetDlgItem(hwnd, cmb3);
    SetMapMode(hDC, MM_ANISOTROPIC);
    switch (ComboBox_GetCurSel(hCmb3))
    {
    case 0:
        SetWindowExtEx(hDC, 100, 100, NULL);
        SetViewportExtEx(hDC, 100, 100, NULL);
        break;
    case 1:
        SetWindowExtEx(hDC, 200, 200, NULL);
        SetViewportExtEx(hDC, 300, 300, NULL);
        break;
    case 2:
        SetWindowExtEx(hDC, 300, 300, NULL);
        SetViewportExtEx(hDC, 200, 200, NULL);
        break;
    }

    HWND hCmb4 = GetDlgItem(hwnd, cmb4);
    switch (ComboBox_GetCurSel(hCmb4))
    {
    case 0:
        SetWindowOrgEx(hDC, -250, -250, NULL);
        break;
    case 1:
        SetWindowOrgEx(hDC, -250 + 80, -250 + 80, NULL);
        break;
    case 2:
        SetWindowOrgEx(hDC, -250 - 80, -250 - 80, NULL);
        break;
    }
    SetViewportOrgEx(hDC, 0, 0, NULL);

    INT x = 0, y = 0;

    MoveToEx(hDC, x - 300, y, NULL);
    LineTo(hDC, x + 300, y);
    MoveToEx(hDC, x, y - 300, NULL);
    LineTo(hDC, x, y + 300);

    if (IsDlgButtonChecked(hwnd, chx1) == BST_CHECKED)
    {
        XFORM xf;
        xf.eM11 = GetDlgItemFloat(hwnd, edt1);
        xf.eM12 = GetDlgItemFloat(hwnd, edt2);
        xf.eM21 = GetDlgItemFloat(hwnd, edt3);
        xf.eM22 = GetDlgItemFloat(hwnd, edt4);
        xf.eDx = GetDlgItemFloat(hwnd, edt5);
        xf.eDy = GetDlgItemFloat(hwnd, edt6);
        SetGraphicsMode(hDC, GM_ADVANCED);
        SetWorldTransform(hDC, &xf);
    }

    LOGFONTA lf;
    ZeroMemory(&lf, sizeof(lf));
    lf.lfHeight = LFHEIGHT;
    lf.lfWidth = LFWIDTH;
    lstrcpyA(lf.lfFaceName, FONTNAME);
    lf.lfOrientation = lf.lfEscapement = GetDlgItemInt(hwnd, cmb1, NULL, TRUE);

    HWND hCmb2 = GetDlgItem(hwnd, cmb2);
    switch (ComboBox_GetCurSel(hCmb2))
    {
    case 0:
        SetTextAlign(hDC, TA_TOP | TA_LEFT);
        break;
    case 1:
        SetTextAlign(hDC, TA_BASELINE | TA_LEFT);
        break;
    case 2:
        SetTextAlign(hDC, TA_BOTTOM | TA_LEFT);
        break;
    }

    SIZE siz;
    if (HFONT hFont = CreateFontIndirectA(&lf))
    {
        HGDIOBJ hFontOld = SelectObject(hDC, hFont);
        {
            SetBkMode(hDC, OPAQUE);
            SetBkColor(hDC, RGB(255, 255, 0));
            SetTextColor(hDC, RGB(0, 0, 0));
            TextOutA(hDC, x, y, g_szText, lstrlen(g_szText));

            GetTextExtentPoint32(hDC, g_szText, lstrlen(g_szText), &siz);
        }
        SelectObject(hDC, hFontOld);
        DeleteObject(hFont);
    }

    SetWindowOrgEx(hDC, 0, 0, NULL);
    ModifyWorldTransform(hDC, NULL, MWT_IDENTITY);
}

#ifndef NO_FREETYPE
    void FreeTypeTextOut(HWND hwnd, HDC hDC, LOGFONTA& lf, TEXTMETRICA tm, POINT pt)
    {
        POINT ptWnd, ptView;
        SIZE sizWnd, sizView;
        GetWindowOrgEx(hDC, &ptWnd);
        GetViewportOrgEx(hDC, &ptView);
        GetWindowExtEx(hDC, &sizWnd);
        GetViewportExtEx(hDC, &sizView);

        SetMapMode(hDC, MM_TEXT);
        SetWindowOrgEx(hDC, 0, 0, NULL);
        SetViewportOrgEx(hDC, 0, 0, NULL);

        FT_Matrix mat = identityMat;

        FT_Matrix matWidth = identityMat;
        IntWidthMatrix(g_face, &matWidth, lf.lfWidth);
        FT_Matrix_Multiply(&matWidth, &mat);

        FT_Matrix matEscape = identityMat;
        lf.lfOrientation = lf.lfEscapement = GetDlgItemInt(hwnd, cmb1, NULL, TRUE);
        IntEscapeMatrix(&matEscape, lf.lfEscapement);
        FT_Matrix_Multiply(&matEscape, &mat);

        FT_Matrix matWorld = identityMat;
        if (IsDlgButtonChecked(hwnd, chx1) == BST_CHECKED)
        {
            ZeroMemory(&matWorld, sizeof(matWorld));
            matWorld.xx = GetDlgItemFloat(hwnd, edt1) * (1 << 16);
            matWorld.yx = -GetDlgItemFloat(hwnd, edt2) * (1 << 16);
            matWorld.xy = -GetDlgItemFloat(hwnd, edt3) * (1 << 16);
            matWorld.yy = GetDlgItemFloat(hwnd, edt4) * (1 << 16);
            FT_Matrix_Multiply(&matWorld, &mat);
        }

        FT_Matrix matLogical = identityMat;
        matLogical.xx = (1 << 16) * sizView.cx / sizWnd.cx;
        matLogical.yx = 0;
        matLogical.xy = 0;
        matLogical.yy = (1 << 16) * sizView.cy / sizWnd.cy;
        FT_Matrix_Multiply(&matLogical, &mat);

        FT_Set_Transform(g_face, &mat, NULL);

        LONGLONG x64, y64;

        FT_Vector vecs[9];
        HWND hCmb2 = GetDlgItem(hwnd, cmb2);
        switch (ComboBox_GetCurSel(hCmb2))
        {
        case 0: // TA_TOP
            vecs[0].x = 0;
            vecs[0].y = 0;                      // lower left
            vecs[1].x = 0;
            vecs[1].y = tm.tmHeight << 16;     // upper left
            vecs[4].x = 0;
            vecs[4].y = tm.tmAscent << 16;      // baseline
            break;
        case 1: // TA_BASELINE
            vecs[0].x = 0;
            vecs[0].y = tm.tmDescent << 16;     // lower left
            vecs[1].x = 0;
            vecs[1].y = -tm.tmAscent << 16;     // upper left
            vecs[4].x = 0;
            vecs[4].y = 0;                      // baseline
            break;
        case 2: // TA_BOTTOM
            vecs[0].x = 0;
            vecs[0].y = -tm.tmHeight << 16;     // lower left
            vecs[1].x = 0;
            vecs[1].y = 0;                      // upper left
            vecs[4].x = 0;
            vecs[4].y = -tm.tmDescent << 16;    // baseline
            break;
        }
        vecs[2] = vecs[1];
        vecs[3] = vecs[0];

        IntEscapeMatrix(&matEscape, -lf.lfEscapement);

        ZeroMemory(&matWorld, sizeof(matWorld));
        matWorld.xx = GetDlgItemFloat(hwnd, edt1) * (1 << 16);
        matWorld.yx = GetDlgItemFloat(hwnd, edt2) * (1 << 16);
        matWorld.xy = GetDlgItemFloat(hwnd, edt3) * (1 << 16);
        matWorld.yy = GetDlgItemFloat(hwnd, edt4) * (1 << 16);
        for (size_t i = 0; i < 9; ++i)
        {
            FT_Vector_Transform(&vecs[i], &matWidth);
            FT_Vector_Transform(&vecs[i], &matEscape);
            if (IsDlgButtonChecked(hwnd, chx1) == BST_CHECKED)
            {
                FT_Vector_Transform(&vecs[i], &matWorld);
            }
            vecs[i].x += GetDlgItemFloat(hwnd, edt5) * (1 << 16);
            vecs[i].y += GetDlgItemFloat(hwnd, edt6) * (1 << 16);
            vecs[i].x += pt.x << 16;
            vecs[i].y += pt.y << 16;
            vecs[i].x >>= 16;
            vecs[i].y >>= 16;
            vecs[i].x = (vecs[i].x - ptWnd.x) * sizView.cx / sizWnd.cx + ptView.x;
            vecs[i].y = (vecs[i].y - ptWnd.y) * sizView.cy / sizWnd.cy + ptView.y;
        }

        x64 = y64 = 0;
        for (size_t i = 0; g_szText[i]; ++i)
        {
            FT_GlyphSlot glyph = LoadGlyphSlot(g_szText[i]);
            x64 += glyph->advance.x;
            y64 -= glyph->advance.y;
        }

        LONGLONG DeltaX64 = x64, DeltaY64 = y64;
        vecs[2].x += DeltaX64 >> 6;
        vecs[2].y += DeltaY64 >> 6;
        vecs[3].x += DeltaX64 >> 6;
        vecs[3].y += DeltaY64 >> 6;

        POINT pts[4];
        for (size_t i = 0; i < 5; ++i)
        {
            pts[i].x = vecs[i].x;
            pts[i].y = vecs[i].y;
        }

        SelectObject(hDC, GetStockObject(WHITE_BRUSH));
        SelectObject(hDC, GetStockObject(BLACK_PEN));
        Polygon(hDC, pts, 4);

        x64 = (vecs[4].x) << 6;
        y64 = (vecs[4].y) << 6;

        HDC hdcMem = CreateCompatibleDC(hDC);
        for (size_t i = 0; g_szText[i]; ++i)
        {
            FT_GlyphSlot glyph = LoadGlyphSlot(g_szText[i]);
            if (HBITMAP hbm = ConvertToGrayscaleBitmap(glyph->bitmap))
            {
                HGDIOBJ hbmOld = SelectObject(hdcMem, hbm);
                {
                    INT rx = (x64 >> 6) + glyph->bitmap_left;
                    INT ry = (y64 >> 6) - glyph->bitmap_top;
                    BitBlt(hDC, rx, ry, glyph->bitmap.width, glyph->bitmap.rows,
                           hdcMem, 0, 0, SRCCOPY);
                }
                SelectObject(hdcMem, hbmOld);
                DeleteObject(hbm);
            }

            x64 += glyph->advance.x;
            y64 -= glyph->advance.y;
        }
        DeleteDC(hdcMem);

        HBRUSH hbr = CreateSolidBrush(RGB(255, 0, 0));
        HGDIOBJ hbrOld = SelectObject(hDC, hbr);
        for (size_t i = 0; i < 5; ++i)
        {
            Ellipse(hDC, pts[i].x - 4, pts[i].y - 4, pts[i].x + 4, pts[i].y + 4);
        }
        SelectObject(hDC, hbrOld);
        DeleteObject(hbr);
    }

    void OnFreeTypeDraw(HWND hwnd, HDC hDC)
    {
        RECT rc;
        GetClientRect(hwnd, &rc);

        HWND hCmb3 = GetDlgItem(hwnd, cmb3);
        SetMapMode(hDC, MM_ANISOTROPIC);
        switch (ComboBox_GetCurSel(hCmb3))
        {
        case 0:
            SetWindowExtEx(hDC, 100, 100, NULL);
            SetViewportExtEx(hDC, 100, 100, NULL);
            break;
        case 1:
            SetWindowExtEx(hDC, 200, 200, NULL);
            SetViewportExtEx(hDC, 300, 300, NULL);
            break;
        case 2:
            SetWindowExtEx(hDC, 300, 300, NULL);
            SetViewportExtEx(hDC, 200, 200, NULL);
            break;
        }

        HWND hCmb4 = GetDlgItem(hwnd, cmb4);
        switch (ComboBox_GetCurSel(hCmb4))
        {
        case 0:
            SetWindowOrgEx(hDC, -250, -250, NULL);
            break;
        case 1:
            SetWindowOrgEx(hDC, -250 + 80, -250 + 80, NULL);
            break;
        case 2:
            SetWindowOrgEx(hDC, -250 - 80, -250 - 80, NULL);
            break;
        }
        SetViewportOrgEx(hDC, 0, 0, NULL);

        INT x = 0, y = 0;

        MoveToEx(hDC, x - 300, y, NULL);
        LineTo(hDC, x + 300, y);
        MoveToEx(hDC, x, y - 300, NULL);
        LineTo(hDC, x, y + 300);

        LOGFONTA lf;
        ZeroMemory(&lf, sizeof(lf));
        lf.lfHeight = LFHEIGHT;
        lf.lfWidth = LFWIDTH;
        lstrcpyA(lf.lfFaceName, FONTNAME);

        TEXTMETRICA tm;
        if (HFONT hFont = CreateFontIndirectA(&lf))
        {
            HGDIOBJ hFontOld = SelectObject(hDC, hFont);
            GetTextMetrics(hDC, &tm);
            SelectObject(hDC, hFontOld);
            DeleteObject(hFont);
        }

        POINT pt = { x, y };
        FreeTypeTextOut(hwnd, hDC, lf, tm, pt);
    }
#endif

void OnPaint(HWND hwnd)
{
    PAINTSTRUCT ps;
    if (HDC hDC = BeginPaint(hwnd, &ps))
    {
#ifndef NO_FREETYPE
        if (IsDlgButtonChecked(hwnd, chx2) == BST_CHECKED)
        {
            OnFreeTypeDraw(hwnd, hDC);
        }
        else
#endif
        {
            OnWinDraw(hwnd, hDC);
        }

        EndPaint(hwnd, &ps);
    }
}

INT_PTR CALLBACK
DialogProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    switch (uMsg)
    {
        HANDLE_MSG(hwnd, WM_INITDIALOG, OnInitDialog);
        HANDLE_MSG(hwnd, WM_COMMAND, OnCommand);
        HANDLE_MSG(hwnd, WM_PAINT, OnPaint);
    }
    return 0;
}

INT WINAPI
WinMain(HINSTANCE   hInstance,
        HINSTANCE   hPrevInstance,
        LPSTR       lpCmdLine,
        INT         nCmdShow)
{
    InitCommonControls();

#ifndef NO_FREETYPE
    if (FT_Error error = FT_Init_FreeType(&g_library))
    {
        MessageBoxA(NULL, "FT_Init_FreeType", NULL, MB_ICONERROR);
        return -1;
    }

    const char *path = GetFontPath(FILENAME);
    if (FT_Error error = FT_New_Face(g_library, path, 0, &g_face))
    {
        MessageBoxA(NULL, "FT_New_Face", NULL, MB_ICONERROR);
        return -2;
    }

    if (FT_Error error = FT_Set_Pixel_Sizes(g_face, 0, -LFHEIGHT))
    {
        MessageBoxA(NULL, "FT_Set_Pixel_Sizes", NULL, MB_ICONERROR);
        return -3;
    }
#endif

    DialogBox(hInstance, MAKEINTRESOURCE(1), NULL, DialogProc);

#ifndef NO_FREETYPE
    FT_Done_Face(g_face);
    FT_Done_FreeType(g_library);
#endif  // ndef NO_FREETYPE

    return 0;
}
